/*
 * disk handler
 *
 * Copyright 2010 Jim Rees jim.rees@rees.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _XOPEN_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>

#include "m68k.h"
#include "dn3000.h"

#define DBLKXTRA 34
#define DBLKDATA 1024
#define DBLKSZ (DBLKXTRA + DBLKDATA)
#define DBLKWDS (DBLKSZ / 2)

/* Maxtor 4380, 382MB */
int disk_ncyl = 1224;
int disk_nhd = 15;
int disk_nsec = 20;

static int fd = -1;

   /*                     7   6   5   4   3   2   1   0                       */
   /*                   +---+---+---+---+---+---+---+---+                     */
   /*                   | 0 | 0 |int|dma|slc| ph|dir|avl|                     */
   /*                   +---+---+---+---+---+---+---+---+                     */

static u_int8_t stb, cstat = 0xc0;
static u_int8_t cmd[10];
static int nblocks; /* number of blocks left to go in current request */

int
disk_init(int dimgfd)
{
    fd = dimgfd;
    lseek(fd, 1024, SEEK_SET);
    return 0;
}

unsigned int
disk_read_memory_8(unsigned int addr)
{
    unsigned int data = 0;

    addr &= 0x3ff;

    if (addr == 0x0) {
	/* check status; everything's hunky-dory of course */
	data = stb;
	cstat = 0xc0;
    } else if (addr == 0x1) {
	/* controller status bits */
	data = cstat;
	cstat = 0xc0;
    } else if (addr == 0x2)
	/* jumpers */
	data = 0x7; /* ESDI */

    printf("disk read %x cstat=%x\n", addr, cstat);

    return data;
}

/*
   win_cmd_blk = packed record case integer of
            0:(cmd_byte    : array [0..5] of char;);
            1:(cmd_word    : array [0..2] of integer;);
            2:(opcode      : 0..255;         { command  }
--
               cyl_msb     : 0..1;           { cylinder MSB }
               lun         : 0..3;           { logical unit number }
               head        : 0..31;          { head number }
--
               cyl_hi      : 0..3;           { not quite high two bits of cylinder }
               sector      : 0..63;          { sector number }
--
               cyl_lo      : 0..255;         { low byte of cylinder address }
--
               block_cnt   : 0..255;         { interleave / block count }
--
               retry_dis   : 0..1;           { retry disable 1=> disabled }
               ecc_dis     : 0..1;           { ECC correction disable 1=> disabled }
               pad1        : 0..7;           { unused }
               step        : 0..7;           { step option }
              );
            end;
*/

unsigned int
disk_read_memory_16(unsigned int addr)
{
    unsigned int data = 0;
    static u_int16_t stbw[4];
    static u_int16_t buf[DBLKWDS];
    static int stbwno, bufi = 0, cyl, hd, sec, blockno;

    if (cmd[0] == 0xec) {
	/* esdi init, return cyls/heads/sec params */
	if (stbwno == 0) {
	    stbw[0] = disk_ncyl - 1;
	    stbw[1] = ((disk_nhd - 1) << 8) | ((disk_nsec - 1) & 0xff);
	    stbw[2] = 0;
	    stbw[3] = 0;
	}
	data = stbw[stbwno++];
	if (stbwno == 4) {
	    stbwno = 0;
	    cstat = 0xcf;
	}
    } else if (cmd[0] == 0x8) {
	/* read data */
	if (bufi == 0) {
	    if (nblocks == 0) {
		cyl = ((cmd[1] & 0x80) << 3) | ((cmd[2] & 0xc0) << 2) | cmd[3];
		hd = cmd[1] & 0x1f;
		sec = cmd[2] & 0x3f;
		nblocks = cmd[4];
		blockno = ((cyl * disk_nhd) + hd) * disk_nsec + sec;
		lseek(fd, blockno * DBLKDATA, SEEK_SET);
		cstat = 0xc9; /* data out phase */
	    }
	    read(fd, &buf[DBLKXTRA / 2], DBLKDATA);
	    nblocks--;
	    printf("disk read %d at %d/%d/%d block %d\n", nblocks, cyl, hd, sec, 0);
	}
	data = buf[bufi++];
	if (bufi >= DBLKWDS - 1) {
	    bufi = 0;
	    if (nblocks <= 0) {
		cstat = 0xcf; /* status phase */
		printf("data out -> status phase\n");
	    }
	}
    } else
	printf("disk read16 %x\n", addr);

    return ntohs(data);
}

void
disk_write_memory_8(unsigned int addr, unsigned int data)
{
    static int cmdi;

    addr &= 0x3ff;

    if (addr == 0x0) {
	cmd[cmdi++] = data;
	if (cmdi == 6) {
	    printf("disk cmd");
	    for (cmdi = 0; cmdi < 6; cmdi++)
		printf(" %x", cmd[cmdi]);
	    printf("\n");
	    cmdi = 0;
	    if (cmd[0] == 0x0)
		cstat = 0xcf;
	    else
		cstat = 0xcb;
	    nblocks = 0;
	}
    } else if (addr == 0x1) {
	cstat = 0xc0;
	printf("disk reset\n");
    } else if (addr == 0x3) {
	cstat = 0xcd;
	printf("slc\n");
    } else
	printf("disk %x: %x\n", addr, data);
}

void
disk_write_memory_16(unsigned int addr, unsigned int data)
{
    unsigned int hdata = htons(data);

    printf("disk write16 %x: %x\n", addr, hdata);
}
