/*
 * Apollo dn3000/3500 emulator
 *
 * Copyright 2010 Jim Rees jim.rees@rees.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef __linux__
#define _XOPEN_SOURCE 600
#define _BSD_SOURCE
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <netdb.h>
#include <fcntl.h>

#include "m68k.h"
#include "dn3000.h"

#define PORT 5900

/* sau7 should be 030, but musashi 3.3 doesn't implement it, need 4.5. */
struct sau {
    char *promname;
    unsigned long phys_base, phys_size;
    unsigned int cpu_type;
} sautab[] = {
    { "DN3000.bin",  0x100000,  0x800000, M68K_CPU_TYPE_68020 },
    { "DN3500.bin", 0x1000000, 0x2000000, M68K_CPU_TYPE_68020 },
};

int bind_socket();
int accept_conn(int sfd);
int repl();
int logline(char *s);

int Nflag, vflag, magic, dflag, hook_tty, sau = 8, quiet, dbgmask;
void *prom_mem, *phys_mem;
unsigned long prom_size, rfc_size, phys_base, phys_size;
int stepcount = 1000000, runstate, vis_width = DISP_VIS_WIDTH;
int use_ring, use_mono = 1, use_color;
char *bindname = "any";
int port = PORT;
static int sleep_msec;
static struct timeval tv_sleepend;

int
main(int ac, char *av[])
{
    int fd, sockfd, vncfd = -1, dimgfd = -1, rv;
    char *s;
    struct stat st;
    struct rfc_head rh;
    int saui, swap_sio = 0;
    char *rfc_name = NULL, *diskimage_fname = NULL;

    while ((s = getav(ac, av)) != NULL) {
	if (!strcmp(s, "-3500"))
	    sau = 7;
	else if (!strcmp(s, "-normal"))
	    Nflag = 1;
	else if (!strcmp(s, "-rfc"))
	    rfc_name = avarg();
	else if (!strncmp(s, "-step", 5))
	    stepcount = atoi(avarg());
	else if (!strcmp(s, "-d"))
	    dflag = 1;
	else if (!strcmp(s, "-v"))
	    vflag = 1;
	else if (!strcmp(s, "-color")) {
	    use_mono = 0;
	    use_color = 1;
	} else if (!strcmp(s, "-ring"))
	    use_ring = 1;
	else if (!strcmp(s, "-sio"))
	    swap_sio = 1;
	else if (!strcmp(s, "-vncv"))
	    vnc_vflag = 1;
	else if (!strcmp(s, "-monov"))
	    mono_vflag = 1;
	else if (!strcmp(s, "-fbv"))
	    fb_vflag = 1;
	else if (!strcmp(s, "-bltv")) {
	    bltv_x = atoi(avarg());
	    bltv_y = atoi(avarg());
	} else if (!strcmp(s, "-keyv"))
	    key_vflag = 1;
	else if (!strcmp(s, "-tty"))
	    hook_tty = 1;
	else if (!strcmp(s, "-unhide"))
	    vis_width = DISP_WIDTH;
	else if (!strcmp(s, "-q"))
	    quiet = 1;
	else if (!strcmp(s, "-bind"))
	    bindname = avarg();
	else if (!strcmp(s, "-port"))
	    port = atoi(avarg());
	else if (!strcmp(s, "-magic"))
	    magic = strtoul(avarg(), NULL, 16);
	else if (!strcmp(s, "-disk"))
	    diskimage_fname = avarg();
    }
    ac -= optind;
    av += optind;

    sockfd = bind_socket();
    if (sockfd < 0)
	exit(1);
    while (1) {
	if (vncfd >= 0)
	    close(vncfd);
	waitpid(-1, NULL, WNOHANG);
	vncfd = accept_conn(sockfd);
	if (!dflag || fork() == 0)
	    break;
    }
    if (vncfd < 0)
	exit(1);

    /* Copy sau params */
    if (sau == 8)
	saui = 0;
    else if (sau == 7)
	saui = 1;
    else {
	fprintf(stderr, "unknown sau%d\n", sau);
	exit(1);
    }
    phys_base = sautab[saui].phys_base;
    phys_size = sautab[saui].phys_size;

    /* Read in prom */
    fd = open(sautab[saui].promname, 0);
    if (fd < 0) {
	fprintf(stderr, "can't open prom\n");
	exit(1);
    }
    fstat(fd, &st);
    prom_size = st.st_size;
    prom_mem = malloc(prom_size);
    read(fd, prom_mem, prom_size);
    close(fd);

    /* Allocate physical memory */
    phys_mem = malloc(phys_size);
    memset(phys_mem, 0, phys_size);

    if (rfc_name) {
	/* Read in rfc file (eventually the prom should do this) */
	fd = open(rfc_name, 0);
	if (fd < 0) {
	    fprintf(stderr, "can't open rfc file %s\n", rfc_name);
	    exit(1);
	}
	fstat(fd, &st);
	rfc_size = st.st_size - sizeof rh;
	read(fd, &rh, sizeof rh);
	read(fd, phys_ptr(ntohl(rh.st)), rfc_size);
	close(fd);
	printf("st=%x en=%x eol=%x ts=%x\n",
	       ntohl(rh.st), ntohl(rh.en), ntohl(rh.eol), ntohl(rh.ts));
    }

    /* Open disk image file */
    if (diskimage_fname != NULL) {
	dimgfd = open(diskimage_fname, 0);
	if (dimgfd < 0)
	    fprintf(stderr, "can't open disk image %s\n", diskimage_fname);
    }

    m68k_set_cpu_type(sautab[saui].cpu_type);
    m68k_init();
    runstate = ST_RESET;

    sio_init(swap_sio, hook_tty);
    if (use_mono)
	mono_init();
    vnc_init(vncfd);
    disk_init(dimgfd);
    ct_init();

    /* The Main Loop */
    rv = repl();

    mono_end();
    sio_end();
    vnc_end();

    free(phys_mem);
    free(prom_mem);

    exit(rv);
}

int
bind_socket()
{
    struct sockaddr_in saddr;
    int i, sfd;

    saddr.sin_family = AF_INET;
    if (!strcmp(bindname, "any"))
	saddr.sin_addr.s_addr = INADDR_ANY;
    else {
	if (!strcmp(bindname, "localhost"))
	    bindname = "127.0.0.1";
	inet_pton(AF_INET, bindname, &saddr.sin_addr);
    }
    saddr.sin_port = htons(port);

    if ((sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	perror("socket");
	return -1;
    }
    i = 1;
    setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &i, sizeof i);

    if (bind(sfd, (struct sockaddr *) &saddr, sizeof saddr) < 0) {
	perror("bind");
	return -1;
    }

    if (listen(sfd, 0) < 0) {
	perror("listen");
	return -1;
    }

    return sfd;
}

int
accept_conn(int sfd)
{
    struct sockaddr_in saddr;
    int fd;
    unsigned int slen;
    time_t t;
    char logstring[100];

    slen = sizeof saddr;

    if ((fd = accept(sfd, (struct sockaddr *) &saddr, &slen)) < 0) {
	perror("accept");
	return -1;
    }

    time(&t);
    sprintf(logstring, "%s C %s", inet_ntoa(saddr.sin_addr), ctime(&t));
    fputs(logstring, stdout);

    if (dflag) {
	logline(logstring);
	quiet = 1;
    }

    return fd;
}

int
repl()
{
    struct timeval tv0, tv1, tv2, tv_cpu, tv_real;
    int nsteps = 0, ostate;
    char logstring[100];

    timerclear(&tv_cpu);
    gettimeofday(&tv0, NULL);

    while (1) {
	if (runstate == ST_RESET) {
	    m68k_pulse_reset();
	    runstate = ST_RUN;
	} else if (runstate == ST_QUIT)
	    break;
	if (runstate == ST_RUN || runstate == ST_STEP) {
	    gettimeofday(&tv1, NULL);
	    m68k_execute(stepcount);
	    gettimeofday(&tv2, NULL);
	    timersub(&tv2, &tv1, &tv2);
	    timeradd(&tv2, &tv_cpu, &tv_cpu);
	    nsteps++;
	}
	if (runstate == ST_STEP)
	    runstate = ST_STOP;
	if (ostate != ST_OFF && runstate == ST_OFF) {
	    led_write_memory_8(0, 0xff);
	    if (use_mono)
		mono_init();
	}
	ostate = runstate;
	sio_run();
	vnc_run(runstate == ST_RUN ? 0 : 100);
	if (sleep_msec) {
	    gettimeofday(&tv1, NULL);
	    if (timercmp(&tv1, &tv_sleepend, >)) {
		runstate = ST_RUN;
		sleep_msec = 0;
	    }
	}
    }
    gettimeofday(&tv_real, NULL);
    timersub(&tv_real, &tv0, &tv_real);

    sprintf(logstring, "%d.%06d sec m68k %d.%06d sec total %.1f MHz\n",
	   (int) tv_cpu.tv_sec, (int) tv_cpu.tv_usec,
	   (int) tv_real.tv_sec, (int) tv_real.tv_usec,
	   (float) nsteps * (float) stepcount / (float) tv_real.tv_sec / 1e6);
    fputs(logstring, stdout);
    if (dflag)
	logline(logstring);

    return 0;
}

void
emu_msleep(int msec)
{
    struct timeval tv0;

    gettimeofday(&tv_sleepend, NULL);
    tv0.tv_sec = msec / 1000;
    tv0.tv_usec = (msec % 1000) * 1000;
    timeradd(&tv_sleepend, &tv0, &tv_sleepend);
    runstate = ST_STOP;
    sleep_msec = msec;
}

int
logline(char *s)
{
    FILE *logf;

    logf = fopen("log", "a");
    if (logf == NULL)
	return -1;

    fputs(s, logf);
    fclose(logf);

    return 0;
}

unsigned int
m68k_read_memory_8(unsigned int addr)
{
    unsigned int data = 0x0;

    if (is_physmem(addr))
	data = *(u_int8_t *) phys_ptr(addr);
    else if (is_monomem(addr))
	data = *(u_int8_t *) disp_ptr(addr);
    else if (addr < prom_size)
	data = *(u_int8_t *) (addr + prom_mem);
    else if ((addr & ~0x3ff) == 0x4d000)
	/* disk */
	data = disk_read_memory_8(addr);
    else if (addr == 0x59000 && use_ring)
	/* ring */
	data = 0x37;
    else if ((addr & ~0x3ff) == 0x50000)
	/* cartridge tape */
	data = ct_read_memory_8(addr);
    else if (addr == 0x5f804)
	/* floppy */
	data = 0x80;
    else if (addr == prom_size + 0x1)
	data = Nflag ? 1 : 0;
    else if ((addr & ~0x1ff) == prom_size + 0x400)
	/* sio line */
	data = sio_read_memory_8(addr);
    else if (is_monoctl(addr) && use_mono)
	/* Mono display */
	data = mono_read_memory_8(addr);
    else if (addr == 0x5e801 && use_color)
	/* Color display type register */
	data = 0x8;
    else if (addr == 0xac00)
	/* someone ran "TE" which is not going to go well */
	runstate = ST_QUIT;
    else if (!quiet)
	printf("ReadByte %x\n", addr);

    return data;
}

unsigned int
m68k_read_memory_16(unsigned int addr)
{
    unsigned int data = 0;

    if (is_physmem(addr))
	data = *(u_int16_t *) phys_ptr(addr);
    else if (is_monomem(addr))
	data = *(u_int16_t *) disp_ptr(addr);
    else if (addr < prom_size)
	data = *(u_int16_t *) (addr + prom_mem);
    else if ((addr & ~0x3ff) == 0x4d000)
	/* disk */
	data = disk_read_memory_16(addr);
    else if (addr == prom_size + 0x1602) {
	/* 3c505 status */
	data = 0xc800;
    } else if (!quiet)
	printf("ReadWord %x\n", addr);

    return ntohs(data);
}

unsigned int
m68k_read_memory_32(unsigned int addr)
{
    unsigned int data = 0;

    if (is_physmem(addr))
	data = *(u_int32_t *) phys_ptr(addr);
    else if (is_monomem(addr))
	data = *(u_int32_t *) disp_ptr(addr);
    else if (addr < prom_size)
	data = *(u_int32_t *) (addr + prom_mem);
    else if (sau != 7 && !quiet)
	printf("ReadLong %x\n", addr);

    return ntohl(data);
}

void
m68k_write_memory_8(unsigned int addr, unsigned int data)
{
    if (is_physmem(addr))
	*(u_int8_t *) phys_ptr(addr) = data;
    else if (is_monomem(addr))
	fb_write_memory_8(addr, data);
    else if ((addr & ~0x3ff) == 0x4d000)
	disk_write_memory_8(addr, data);
    else if (is_monoctl(addr))
	mono_write_memory_8(addr, data);
    else if ((addr & ~0x3ff) == 0x50000)
	ct_write_memory_8(addr, data);
    else if (((addr - prom_size) & ~0x1ff) == 0x1400)
	pic_write_memory_8(addr, data);
    else if ((addr & ~0x1ff) == prom_size + 0x400)
	sio_write_memory_8(addr, data);
    else if (addr == prom_size + 0x100) {
	led_write_memory_8(addr, data);
    } else if (!quiet)
	printf("WriteByte %x: %x\n", addr, data);
}

void
m68k_write_memory_16(unsigned int addr, unsigned int data)
{
    unsigned int hdata = htons(data);

    if (is_physmem(addr))
	*(u_int16_t *) phys_ptr(addr) = hdata;
    else if (is_monomem(addr))
	fb_write_memory_16(addr, data);
    else if ((addr & ~0x3ff) == 0x4d000)
	disk_write_memory_16(addr, data);
    else if (is_monoctl(addr))
	mono_write_memory_16(addr, data);
    else if (hdata != 0 && !quiet)
	printf("WriteWord %x: %x\n", addr, hdata);
}

void
m68k_write_memory_32(unsigned int addr, unsigned int data)
{
    unsigned int hdata = htonl(data);

    if (is_physmem(addr))
	*(u_int32_t *) phys_ptr(addr) = hdata;
    else if (is_monomem(addr))
	fb_write_memory_32(addr, data);
    else if (!quiet)
	printf("WriteLong %x: %x\n", addr, hdata);
}

void
pic_write_memory_8(unsigned int addr, unsigned int data)
{
    int picno = (addr & 0x100) >> 8;
    static u_int8_t picreg[2];

    if (picreg[picno] != 0xff) {
	if (data == 0xff)
	    printf("PIC %d mask all interrupts\n", picno);
	picreg[picno] = data;
    } else
	printf("PIC %d: %x\n", picno, data);
}
