/*
 * vnc server
 *
 * Copyright 2010 Jim Rees jim.rees@rees.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>

typedef u_int8_t CARD8;
typedef u_int16_t CARD16;
typedef u_int32_t CARD32;
typedef u_int16_t pinteger;

#include "dn3000.h"
#include "rfbproto.h"

#define roundup(x, y) ((x) - ((x) % (y)) + (y))
#define MIN(x, y) ((x) < (y) ? (x) : (y))

#define RS 16	/* Rectangle Size; all rects are RSxRS */

extern u_char monotab[256][8];
extern u_char stipple[], led_bkg[], led_pwr[], led_on[], led_off[];

char *clmsgnames[] = {
    "rfbSetPixelFormat",
    "rfbFixColourMapEntries",
    "rfbSetEncodings",
    "rfbFramebufferUpdateRequest",
    "rfbKeyEvent",
    "rfbPointerEvent",
    "rfbClientCutText",
};

void proto_init(FILE *f);
int update(FILE *f, rfbClientToServerMsg *mp);
void sendrhd(FILE *f, int x, int y, int enc);
void sendrect(FILE *f, u_char *p, int lw);
void sendcopy(FILE *f, int x, int y);
void sendsolid(FILE *f, u_char px);
int ispx(u_char *p, int lw, pinteger px);
int rectcmp(u_char *p0, int lw0, u_char *p1, int lw1);
void sendkey(int down, unsigned int key);
void client_cut(FILE *f, char *s);
int netread(void *p, int elemsize, int numelem, FILE *f);

int vnc_vflag;
static FILE *f;
static int fd, update_req;
static int willCopyRect, willRRE;
unsigned int leds = 0xff;

int
vnc_init(int vncfd)
{
    f = fdopen(vncfd, "w");
    setbuffer(f, (char *) malloc(16 * 1024), 16 * 1024);
    proto_init(f);
    fd = vncfd;

    return 0;
}

void
proto_init(FILE *f)
{
    int n;
    CARD32 auth;
    rfbProtocolVersionMsg pvmsg;
    rfbClientInitMsg clmsg;
    rfbServerInitMsg msg;
    char hostname[100];

    fprintf(f, rfbProtocolVersionFormat, rfbProtocolMajorVersion, rfbProtocolMinorVersion);
    fflush(f);
    n = read(fileno(f), pvmsg, sz_rfbProtocolVersionMsg);
    pvmsg[n] = '\0';
    if (vnc_vflag)
	printf("client version %s", pvmsg);

    auth = htonl(rfbNoAuth);
    fwrite(&auth, sizeof auth, 1, f);
    fflush(f);

    netread(&clmsg, sz_rfbClientInitMsg, 1, f);

    msg.framebufferWidth = htons(vis_width);
    msg.framebufferHeight = htons(DISP_HEIGHT);
    msg.format.bitsPerPixel = 8;
    msg.format.depth = 8;
    msg.format.bigEndian = 1;
    msg.format.trueColour = 1;
    msg.format.redMax = htons(7);
    msg.format.greenMax = htons(7);
    msg.format.blueMax = htons(3);
    msg.format.redShift = 0;
    msg.format.greenShift = 3;
    msg.format.blueShift = 6;
    msg.format.pad1 = 0;
    msg.format.pad2 = htons(0);
    gethostname(hostname, sizeof hostname);
    msg.nameLength = htonl(strlen(hostname));
    fwrite(&msg, sz_rfbServerInitMsg, 1, f);
    fputs(hostname, f);
    fflush(f);

    if (vnc_vflag)
	printf("proto_init done\n");
}

void
vnc_run(int block)
{
    int i;
    fd_set fdset;
    struct timeval tv;
    static rfbClientToServerMsg msg, upd_msg;
    CARD32 enc;
    char *cp;

    if (f == NULL)
	return;

    if (update_req) {
	/*
	 * Client asked for an update previously but we had nothing for him
	 * at the time (update_req=1).
	 */
	if (update(f, &upd_msg))
	    update_req = 0;
    }

    FD_ZERO(&fdset);
    FD_SET(fd, &fdset);
    tv.tv_sec = 0;
    tv.tv_usec = block * 1000;
    if (select(fd + 1, &fdset, NULL, NULL, &tv) < 0) {
	runstate = ST_QUIT;
	return;
    }
    if (!FD_ISSET(fd, &fdset))
	return;

    if (netread(&msg.type, sizeof msg.type, 1, f) != 1) {
	runstate = ST_QUIT;
	return;
    }

    if (vnc_vflag && msg.type != rfbPointerEvent)
	printf("client message %s\n", clmsgnames[msg.type]);

    switch(msg.type) {
    case rfbSetPixelFormat:
	netread(&msg.type + 1, sz_rfbSetPixelFormatMsg - 1, 1, f);
	if (msg.spf.format.bitsPerPixel != 8) {
	    printf("vncd: client pixel format incompatible\n");
	    exit(1);
	}
	break;
    case rfbFixColourMapEntries:
	netread(&msg.type + 1, sz_rfbFixColourMapEntriesMsg - 1, 1, f);
	break;
    case rfbSetEncodings:
	netread(&msg.type + 1, sz_rfbSetEncodingsMsg - 1, 1, f);
	for (i = 0; i < ntohs(msg.se.nEncodings); i++) {
	    netread(&enc, sizeof enc, 1, f);
	    switch(ntohl(enc)) {
	    case rfbEncodingCopyRect:
		willCopyRect = 1;
		break;
	    case rfbEncodingRRE:
		willRRE = 1;
		break;
	    }
	}
	break;
    case rfbFramebufferUpdateRequest:
	netread(&msg.type + 1, sz_rfbFramebufferUpdateRequestMsg - 1, 1, f);
	if (!update(f, &msg)) {
	    upd_msg = msg;
	    update_req = 1;
	}
	break;
    case rfbKeyEvent:
	netread(&msg.type + 1, sz_rfbKeyEventMsg - 1, 1, f);
	sendkey(msg.ke.down, ntohl(msg.ke.key));
	break;
    case rfbPointerEvent:
	netread(&msg.type + 1, sz_rfbPointerEventMsg - 1, 1, f);
	break;
    case rfbClientCutText:
	netread(&msg.type + 1, sz_rfbClientCutTextMsg - 1, 1, f);
	i = ntohl(msg.cct.length);
	cp = (char *) malloc(i+1);
	netread(cp, 1, i, f);
	cp[i] = '\0';
	client_cut(f, cp);
	free(cp);
	break;
    }

    fflush(f);
}

#define ledc(x) ((x) ? ' ' : '*')

void
led_write_memory_8(unsigned int addr, unsigned int data)
{
    if (data != leds) {
	leds = (leds & ~0xff) | data;
#ifdef PRINTLEDS
	printf("%c %c %c %c\n",
	       ledc(leds & 0x8), ledc(leds & 0x4), ledc(leds & 0x2), ledc(leds & 0x1));
#endif
    }
}

void
vnc_bell()
{
    rfbBellMsg m;

    m.type = rfbBell;
    fwrite(&m, sizeof m, 1, f);
    fflush(f);
}

void
vnc_end()
{
    if (f) {
	fclose(f);
	f = NULL;
    }
}

int
update(FILE *f, rfbClientToServerMsg *mp)
{
    int x, y, x0, y0, w, h, x1, y1, r, xstip = -1, ystip;
    int inc, ncpy = 0, nwht = 0, nblk = 0, nraw = 0, nrect = 0;
    rfbFramebufferUpdateMsg umsg;
    u_char *p, *r1;
    static u_char *p0, *p1;
    short lw;
    static int rlist[((DISP_HEIGHT + RS - 1) / RS * (DISP_WIDTH + RS - 1) / RS) + 100];
    int rno, *rlistp;
    static int oleds = -1;

    inc = mp->fur.incremental;
    x0 = ntohs(mp->fur.x);
    y0 = ntohs(mp->fur.y);
    w = ntohs(mp->fur.w);
    h = ntohs(mp->fur.h);

    r = x0 % RS;
    x0 -= r;
    w = roundup(w + r, RS);
    x1 = MIN(x0 + w, vis_width);
    r = y0 % RS;
    y0 -= r;
    h = roundup(h + r, RS);
    y1 = MIN(y0 + h, DISP_HEIGHT);

    /* Put previous frame in p0, current frame in p1 */
    p = p0;
    p0 = p1;
    p1 = p;

#ifdef USE_GPR
    gpr_$inq_bitmap_pointer(bmdesc, (char **) &p, &lw, &st);
    if (st.all != status_$ok) {
	ErrPrint("gpr_$inq_bitmap_pointer", st);
	gpr_$terminate(false, &st);
	return 0;
    }
    lw *= 2;
#else
    p = disp_mem;
    lw = DISP_WIDTH / 8;
#endif

    if (p1 == NULL)
	p1 = (u_char *) malloc(DISP_HEIGHT * lw);
    memmove(p1, p, DISP_HEIGHT * lw);

    /*
     * p0, p1 are the previous and current frame buffers.
     * We scan these frames looking for changed rectangles.
     * r1 is the rectangles under consideration in p1.
     * rlist is the list of rectangles we're going to send, nrect is the count.
     * lw is the width of a line in bytes.
     */

    rlistp = rlist;
    for (y = y0; y < y1; y += RS) {
	for (x = x0; x < x1; x += RS) {
	    rno = y * lw + x / 8;
	    if (!p0 || !inc) {
		/* can't do incremental, put all rects on list */
		*rlistp++ = rno;
	    } else if (x >= 800 && x < 880 && y == DISP_HEIGHT - RS) {
		/* LEDs */
		if (leds != oleds)
		    *rlistp++ = rno;
	    } else if (rectcmp(&p0[rno], lw, &p1[rno], lw))
		/* rectangle has changed, put it on the list */
		*rlistp++ = rno;
	}
    }
    *rlistp = -1;
    nrect = rlistp - rlist;

    if (nrect == 0)
	return 0;

    if (vnc_vflag)
	printf("%d rectangles in %d %d %d %d %s\n", nrect, x0, y0, w, h, inc ? "inc" : "full");

    /* Now that we have the list, send the rectangles. */

    umsg.type = rfbFramebufferUpdate;
    umsg.pad = 0;
    umsg.nRects = htons(nrect);
    fwrite(&umsg, sizeof umsg, 1, f);

    for (rlistp = rlist; *rlistp != -1; rlistp++) {
	rno = *rlistp;
	x = (rno % lw) * 8;
	y = rno / lw;
	r1 = &p1[rno];

	if (x >= 800 && x < 880 && y == DISP_HEIGHT - RS) {
	    /* Paste in the LEDs in a crude way */
	    int ledno = (x - 800) / 16;
	    sendrhd(f, x, y, rfbEncodingRaw);
	    if (ledno <= 3)
		fwrite((leds & (1 << (3 - ledno))) ? led_off : led_on, 1, 256, f);
	    else if (ledno == 4)
		fwrite(runstate == ST_OFF ? led_off : led_pwr, 1, 256, f);
	    nraw++;
	} else if (willCopyRect && !rectcmp(r1, lw, stipple, RS / 8)) {
	    if (xstip < 0) {
		/* no stipple sent yet; send one and remember where */
		sendrhd(f, x, y, rfbEncodingRaw);
		sendrect(f, r1, lw);
		xstip = x;
		ystip = y;
		nraw++;
	    } else {
		/* stipple already sent; copy it */
		sendrhd(f, x, y, rfbEncodingCopyRect);
		sendcopy(f, xstip, ystip);
		ncpy++;
	    }
	} else if (willRRE && ispx(r1, lw, 0x0000)) {
	    sendrhd(f, x, y, rfbEncodingRRE);
	    sendsolid(f, 0x00);
	    nblk++;
	} else if (willRRE && ispx(r1, lw, 0xffff)) {
	    sendrhd(f, x, y, rfbEncodingRRE);
	    sendsolid(f, 0xff);
	    nwht++;
	} else {
	    sendrhd(f, x, y, rfbEncodingRaw);
	    sendrect(f, r1, lw);
	    nraw++;
	}
    }

    if (vnc_vflag)
	printf("%d copy, %d black, %d white, %d raw\n", ncpy, nblk, nwht, nraw);

    oleds = leds;
    fflush(f);

    return 1;
}

void
sendrhd(FILE *f, int x, int y, int enc)
{
    rfbFramebufferUpdateRectHeader rhd;

    rhd.r.x = htons(x);
    rhd.r.y = htons(y);
    rhd.r.w = htons(RS);
    rhd.r.h = htons(RS);
    rhd.encoding = htonl(enc);
    fwrite(&rhd, sizeof rhd, 1, f);
}

void
sendrect(FILE *f, u_char *p, int lw)
{
    int x, y;

    for (y = 0; y < RS; y++)
	for (x = 0; x < RS / 8; x++)
	    fwrite(monotab[p[y * lw + x]], 1, 8, f);
}

void
sendcopy(FILE *f, int x, int y)
{
    rfbCopyRect rcr;

    rcr.srcX = htons(x);
    rcr.srcY = htons(y);
    fwrite(&rcr, sz_rfbCopyRect, 1, f);
}

void
sendsolid(FILE *f, u_char px)
{
    struct {
	rfbRREHeader rre;
	CARD8 px;
    } h;

    h.rre.nSubrects = htonl(0);
    h.px = px;
    fwrite(&h, sz_rfbRREHeader + 1, 1, f);
}

/*
 * ispx and rectcomp have been optimized for RS=16.
 * They won't work for other values.
 */
int
ispx(u_char *p, int lw, pinteger px)
{
    int y;

    for (y = 0; y < RS; y++)
	if (*(pinteger *)(p + (y * lw)) != px)
	    return 0;
    return 1;
}

int
rectcmp(u_char *p0, int lw0, u_char *p1, int lw1)
{
    int y;

    if (((pinteger *) p0)[0] != ((pinteger *) p1)[0])
	return 1;
    for (y = 0; y < RS; y++)
	if (((pinteger *) p0)[y * lw0 / 2] != ((pinteger *) p1)[y * lw1 / 2])
	    return 1;
    return 0;
}

void
sendkey(int down, unsigned int key)
{
    int i;
    static int mods;

    if ((key & ~0xf) == 0xffe0) {
	/* modifiers */
	i = key & 0xf;
	if (i == 2) {
	    /* map right-shift to left-shift */
	    i = 1;
	}
	i <<= 16;
	if (down)
	    mods |= i;
	else
	    mods &= ~i;
	return;
    }
    if (!down)
	return;

    if (key >= 0xff00 && key < 0xff20) {
	/* Control char */
	key &= ~0xff00;
    }

    if (!quiet)
	printf("sendkey: %x\n", key | mods);

    if (key <= '~' && (mods & ~0x10000) == 0) {
	sio_inskey(key);
    } else {
	key |= mods;
	if (key == 0x3000d || key == 0x40071) {
	    /* ctl-ret, rt-ctl-q */
	    runstate = ST_QUIT;
	} else if (key == 0x40068) {
	    /* rt-ctl-h */
	    if (runstate == ST_OFF)
		runstate = ST_RESET;
	    else
		runstate = ST_OFF;
	    leds ^= 0x100;
	} else if (key == 0x40070) {
	    /* rt-ctl-p */
	    if (runstate == ST_STOP)
		runstate = ST_RUN;
	    else if (runstate == ST_RUN)
		runstate = ST_STOP;
	} else if (key == 0x40072) {
	    /* rt-ctl-r */
	    runstate = ST_RESET;
	} else if (key == 0x40073) {
	    /* rt-ctl-s */
	    if (runstate == ST_STOP || runstate == ST_RUN)
		runstate = ST_STEP;
	} else if (key == 0x40074) {
	    /* rt-ctl-t */
	    magic++;
	} else if (key >= 0x40000 + '0' && key <= 0x40000 + '9') {
	    dbgmask ^= (1 << (key & 0xf));
	    printf("dbgmask=%x\n", dbgmask);
	} else if (key == 0x0ff61) {
	    /* print screen */
	    mono_print_screen(0, 0, vis_width, DISP_HEIGHT);
	}
    }
}

void
client_cut(FILE *f, char *s)
{
    FILE *pf;

    pf = fopen("`node_data/paste_buffers/default.txt", "w");
    if (pf == NULL)
	return;
    fputs(s, pf);
    fclose(pf);
}

int
netread(void *p, int elemsize, int numelem, FILE *f)
{
    int n, len = elemsize * numelem;
    u_char *cp = p;

    while (len > 0) {
	n = read(fileno(f), cp, len);
	if (n <= 0)
	    break;
	cp += n;
	len -= n;
    }
    return ((elemsize * numelem - len) / elemsize);
}
